define get_diverging(t)
{
   %% cold hot scheme
   variable R = 0.237 - 2.13*t + 26.92*t*t - 65.5*t*t*t + 63.5*t^4 - 22.36*t^5 ;
   variable G = ((0.572 + 1.524*t -1.811*t*t)/(1. - 0.291*t + 0.1574*t*t))^2 ;
   variable B = 1./(1.579 - 4.03*t + 12.92*t*t  - 31.4*t*t*t + 48.6*t^4 -23.36*t^5) ;
   return rgb2hex(R,G,B ; str) ;
}

private define get_rainbow (t)
{
   %% rainbow scheme
   variable R = (0.472 - 0.567*t + 4.05*t*t)/(1 + 8.72*t - 19.17*t*t + 14.1*t*t*t)  ;
   variable G = (0.108932 - 1.22635*t + 27.284*t*t - 98.577*t*t*t  + 163.3*t^4  - 131.395*t^5 + 40.634*t^6) ;
   variable B = 1./(1.97 + 3.54*t - 68.5*t*t + 243*t*t*t - 297*t^4 + 125*t^5) ;
   return rgb2hex(R,G,B ; str) ;
}

%%%%%%%%%%%%%%%%%%%%%%%
define get_sron_colors()
%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{get_sron_colors}
%\synopsis{get colors for different data, with good color blind readability}
%\usage{String_Type colors = get_sron_colors(Integer_Type number)}
%\altusage{Struct_Type colors = get_sron_colors(Integer_Type number ; names)}
%\altusage{String_Type color = get_sron_colors(Double_Type value ; cont)}
%
%\description
%    The color schemes in this function are based on SRON Technical
%    note SRON/EPS/TN/09-002 (Issue 2.2 from 2012-12-29): "Colour
%    Schemes" prepared by Paul Tol. Three color schemes are avaible:
%     - a quantitative (default): up to 12 distinct colors, not sorted
%     (based on Fig. 3)
%     - a diverging continuous one, running from blue to red (Fig. 8)
%     - a continouos rainbow one (Fig. 13)
%     
%     Selection of the color schemes is done with the qualifier "type"
%     ("qualitative" (default), "diverging" or "rainbow")
%     
%     The function returns the hex-code strings for each color.
%     For the quantitative scheme it is also possible to return a
%     structure with colornames and hex value (qualifier "names").
%     
%     For the continouous schemes it is also possible to get the color
%     corresponding to a (float) value between 0 and 1 (qualifier
%     "cont").
%
%\qualifiers{
%\qualifier{type}{=String_Type select color scheme ("qualitative" (default),"diverging", or "rainbow"}
%\qualifier{names}{(Boolean) return color names in structure for qualitative color scheme}
%\qualifier{cont}{(Boolean) change meaning of argument to value between 0 and 1, return corresponding color}
%}
%!%-
{
   variable n ;
   switch(_NARGS)
     { case 1: n = () ;}
     { help(_function_name()); return; }

   variable type = qualifier("type", "qualitative") ;

   if (qualifier_exists("cont") == 0 and typeof(n) != Integer_Type)
     {
	()=printf("Error (get_srong_colors): Non Integer input only allowed with \"cont\" qualifier.\n", type);
	return ;
     }
   
   if (type == "qualitative")
     {
	if (qualifier_exists("cont"))
	  {
	     ()=printf("Error (get_sron_colors): cont qualifier not compatible with selected color scheme (%s).\n", type);
	     return ;
	  }
	
	variable cols = [
			 "#4477AA", 
			 "#CC6677", 
			 "#DDCC77", 
			 "#117733", 
			 "#332288", 
			 "#88CCEE", 
			 "#AA4499", 
			 "#44AA99", 
			 "#999933", 
			 "#882255", 
			 "#661100", 
			 "#6699CC", 
			 "#AA4466", 
			];
   
	variable names = [
			  "blue1", 
			  "red1",  
			  "yellow1",
			  "green1",
			  "purple1",
			  "blue2", 
			  "pink1", 
			  "green2",
			  "green3",
			  "red2",  
			  "red3",  
			  "blue3", 
			  "red4"
			 ];
      
	variable ndx ;
   
	switch (n)
	  { n == 1 : ndx = [0] ; }
	  { n == 2 : ndx = [0,1] ; }
	  { n == 3 : ndx = [0,2,1] ; }
	  { n == 4 : ndx = [0,3,2,1] ; }
	  { n == 5 : ndx = [4,5,3,2,1] ; }
	  { n == 6 : ndx = [4,5,3,2,1,6] ; }
	  { n == 7 : ndx = [4,5,7,3,2,1,6] ; }
	  { n == 8 : ndx = [4,5,7,3,8,2,1,6] ; }
	  { n == 9 : ndx = [4,5,7,3,8,2,1,9,6] ; }
	  { n == 10 : ndx = [4,5,7,3,8,2,10,1,9,6] ; }
	  { n == 11 : ndx = [4,11,5,7,3,8,2,10,1,9,6] ; }
	  { n == 12 : ndx = [4,11,5,7,3,8,2,10,1,12,9,6] ; }
	  {
	     ()=printf("Error (get_srong_colors): requested more than 12 colors for qualitative color schme!\n");
	     return;
	  }

	variable i ;
   
	if (qualifier_exists("names"))
	  {
	     variable mystruct = @Struct_Type(names[ndx]) ;
	     _for i (0, length(ndx)-1, 1)
	       {
		  set_struct_field(mystruct, names[ndx[i]], cols[ndx[i]]) ;
	       }
	     return mystruct ;
	  }

	return cols[ndx] ;
     }
   
   if (type == "diverging" or type == "rainbow")
     {
	if (qualifier_exists("names"))
	  {
	     ()=printf("Warning (get_sron_colors): names not available with %s color scheme, \
will return color values only!\n", qualifier("type"));
	  }

	variable colfun ;
	
	if (type == "diverging")
	  {
	     colfun = &get_diverging ;
	  }

	if (type == "rainbow")
	  {
	     colfun = &get_rainbow ;
	  }

	if (qualifier_exists("cont"))
	  {
	     ifnot (min(n) >= 0 and max(n) <= 1)
	       {
		  ()=printf("Error (get_sron_colors): for continuous colors, argument needs to be between 0 and 1!\n");
		  return ;
	       }
	     
	     return @colfun(n) ;
	  }

	return array_map(String_Type, colfun , [0:1:#n]) ;
     }
   
   else 
     {
	()=printf("Error (get_sron_colors): Unknown color scheme type (%s) requested!\n", type);
	return ;
     }
   
}

      
