define BinaryCor()
%!%+
%\function{BinaryCor}
%\synopsis{Removes the influence of the doublestar motion for circular or eliptical orbits.}
%\usage{Array_Type t = BinaryCor(Array_Type OR Double_Type time), time in MJD}
%\qualifiers{
%\qualifier{asini}{Projected semi-major axis [lt-secs], Mandatory}
%\qualifier{porb}{Orbital period at the epoch [days], Mandatory}
%\qualifier{eccentricity}{Eccentricity, (0<=e<1)}
%\qualifier{omega}{Longitude of periastron [degrees], mandatory}
%\qualifier{t0}{epoch for mean longitude of 0 degrees (periastron, MJD)}
%\qualifier{t90}{epoch for mean longitude of 90 degrees (MJD)}
%\qualifier{pporb}{rate of change of the orbital period (s/s) (default 0)}
%\qualifier{limit}{absolute precision of the correction of the  computation (days, default: 1D-6)}
%\qualifier{maxiter}{stop Newton-Raphson iteration after maxiter steps if limit is not reached (default: 20)}
%}
%\description
%    (transcribed from IDL-programm BinaryCor.pro)
%
%    For each time, the z-position of the emitting object is computed
%    and the time is adjusted accordingly. This is iterated until
%    convergence is reached (usually only one iteration is necessary,
%    even in high elliptic cases).
%
%    Follows equations from Hilditch's book and has also been
%    checked against fasebin/axBary. All codes give identical results,
%    (to better than 1d-7s) as checked by a Monte Carlo search using
%    1d7 different orbits.
%
%    qualifiers t90 and t0 have to be in days and in the same time
%    system as time (e.g. JD or MJD)
%
%    Circular orbits:
%         * if time of lower conjunction Tlow is known, set
%           t0=Tlow and omega=0
%         * if time of ascending node is known, Tasc, set
%           t90=Tasc and omega=0
%         * if time of mid eclipse is known, Tecl, set
%           t0=Tecl-0.25*porb and omega=0
%!%-
{
  variable time;
  switch(_NARGS)
  { case 1: time = (); }
  { help(_function_name()); return; }
  %%make sure that time is an array
  time = [time];

  %% qualifiers
  variable asini   = qualifier("asini");
  variable porb    = qualifier("porb");
  variable ecc     = qualifier("eccentricity");
  variable omega_d = qualifier("omega");
  variable t90     = qualifier("t90");
  variable t0      = qualifier("t0");
  variable pporb   = qualifier("pporb",0.0);
  variable limit   = qualifier("limit",1e-6);
  variable maxiter = qualifier("maxiter",20);

  %% check if required qualifiers assini, poen and eccentricity
  %% are assigned, die else
  ifnot(qualifier_exists("asini"))
  { message("error: asini value not assigned");
    return;
  }

  ifnot(qualifier_exists("porb"))
  { message("error: porb value not assigned");
    return;
  }

  ifnot(qualifier_exists("eccentricity"))
  { message("error: eccentricity value not assigned");
    return;
  }

  %% look that one and only one of the t0 and t90 values is assigned
  ifnot(qualifier_exists("t0") or qualifier_exists("t90"))
  { message("error: need t90 or t0 value");
    return;
  }

  if(qualifier_exists("t90")  and qualifier_exists("t0"))
  { message("error: Only one of the t90 and t0 arguments is allowed");
    return;
  }

  %% check that omega is set for noncircular orbits
  if(ecc != 0)
    ifnot(qualifier_exists("omega"))
    { message("error: Need omega keyword for noncircular orbits");
      return;
    }
  else  %% if omega not-set for circular orbits, set to 0.
    ifnot(qualifier_exists("omega"))
      omega_d=0.;

  %%check if eccentricity is 0<=ecc<1
  if(ecc < 0)
  { message("error: eccentricity must be positive!");
    return;
  }
  if(ecc >=1)
  { message("error: Orbit correction has only been implemented for circular and elliptic orbits");
    return;
  }

  %% compute t0 from t90 if necessary
  %% note that this assumes pporb<<porb, which can be generally assumed
  if(qualifier_exists("t90"))
   t0 = t90+(omega_d-90.)/360. * porb;

  %% set default values
  if (maxiter <=0)
    maxiter=20;

  variable asini_d = double(asini)/86400.;
  variable t = time;

  %%Corrections for eccentricity 0<=ecc<1
  variable omega = omega_d * PI/180.0;
  variable sinw = sin(omega);
  variable cosw = cos(omega);
  variable sq = sqrt((1.-ecc)*(1.+ecc));
  variable cor =2.*limit;

  %%start with number of iterations = zero
  variable numiter=0;
  while((abs(max(cor)) > limit) and (numiter < maxiter))
  {
    variable tper = (t-t0)/porb;
    variable m = 2*PI*(tper*(1.-0.5*pporb*tper));
    variable eanom = KeplerEquation(m,ecc);
    variable sin_e = sin(eanom);
    variable cos_e = cos(eanom);
    variable z = asini_d*(sinw*(cos_e-ecc)+sq*cosw*sin_e);
    variable f = (t-time)+z;
    variable df = (sq*cosw*cos_e - sinw*sin_e)*(2*PI*asini_d/(porb*(double(1.0)-ecc*cos_e)));
    cor = f/(double(1)+df);
    t -= cor;
    numiter++;
  }

  if(numiter >= maxiter)
    message("Exceeded maxiter iterations and did not reach convergence");
  return t;
}
