define cel2gal() %{{{
%!%+
%\function{cel2gal}
%\synopsis{Transform celestial to Cartesian Galactic coordinates}
%\usage{cel2gal(Double_Types ah[], am[], as[], dd[], dm[], ds[], dist[], vrad[], pma_cos_d[], pmd[]; qualifiers)}
%\description
%    Transform celestial coordinates (right ascension [h, m, s], declination [deg, arcmin, arcsec],
%    distance [kpc], radial velocity [km/s], proper motion in right ascension times cosine of declination
%    [mas/yr], proper motion in declination [mas/yr]) to right-handed, Cartesian Galactic coordinates
%    (x [kpc], y [kpc], z [kpc], vx [km/s], vy [km/s], vz [km/s]) with the Galactic center at the origin,
%    the Sun on the negative x-axis and the z-axis pointing to the north Galactic pole implying clockwise
%    Galactic rotation when seen from the half space with positive z. See function 'RD2rad' for detailed
%    information on how to properly enter right ascension and declination.
%\qualifiers{
%\qualifier{parallax}{Set this qualifier to use parallaxes [mas] instead of distances [kpc].}
%\qualifier{SunGCDist}{Sun-Galactic center distance [kpc];
%      default: 8.4 (see Model I in Irrgang et al., 2013, A&A, 549, A137)}
%\qualifier{vxs}{Sun's x-velocity component [km/s] relative to the local standard of rest;
%      default: 11.1 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vys}{Sun's y-velocity component [km/s] relative to the local standard of rest;
%      default: 12.24 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vzs}{Sun's z-velocity component [km/s] relative to the local standard of rest;
%      default: 7.25 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vlsr}{Local standard of rest velocity [km/s];
%      default: 242 (see Model I in Irrgang et al., 2013, A&A, 549, A137)}
%\qualifier{GC_NGP}{Celestial coordinates of the Galactic center and the north Galactic pole in the
%      format [ah_GC, am_GC, as_GC, dd_GC, dm_GC, ds_GC, ah_NGP, am_NGP, as_NGP, dd_NGP, dm_NGP, ds_NGP];
%      default: [17, 45, 37.224, -28, 56, 10.23, 12, 51, 26.282, 27, 07, 42.01] (J2000, Reid & Brunthaler, 2004, ApJ, 616, 872)}
%}
%\example
%    (x, y, z, vx, vy, vz) = cel2gal(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
%    % -> Sun's position and velocity
%    (x, y, z, vx, vy, vz) = cel2gal(17, 45, 37.224, -28, 56, 10.23, 8.4, -11.1, -3.171, -5.544);
%    % -> Galactic center's position and velocity
%    (x, y, z, vx, vy, vz) = cel2gal([0,17], [0,45], [0,37.224], [0,-28], [0,56], [0,10.23], [0,8.4], [0,-11.1], [0,-3.171], [0,-5.544]);
%    % -> position and velocity of Sun and Galactic center stored in the same array
%    (x, y, z, vx, vy, vz) = cel2gal(17, 45, 37.224, -28, 56, 10.23, 8.4, -11.1, -3.171, -5.544;
%    GC_NGP=[17, 45, 37.1991, -28, 56, 10.221, 12, 51, 26.2755, 27, 7, 41.704]);
%    % -> use non-standard coordinates for Galactic center and north Galactic pole
%    (x, y, z, vx, vy, vz) = cel2gal( gal2cel(0, 0, 0, 0, 0, 0) );
%    % compute vr and vphi:
%    vr = (x*vx+y*vy)/sqrt(x^2+y^2);
%    vphi = (y*vx-x*vy)/sqrt(x^2+y^2);
%\seealso{RD2rad, gal2cel}
%!%-
{
  if(_NARGS!=10)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable ah, am, as, dd, dm, ds, dist, rv, pma, pmd;
    (ah, am, as, dd, dm, ds, dist, rv, pma, pmd)=();
    ah = [1.*ah]; am = [1.*am]; as = [1.*as]; dd = [1.*dd]; dm = [1.*dm]; ds = [1.*ds]; dist = [1.*dist]; rv = [1.*rv]; pma = [1.*pma]; pmd = [1.*pmd];
    if(qualifier_exists("parallax"))
      dist = 1./dist; % conversion from parallax in units of mas to distance in kpc
    variable ind = where(dist<0);
    if(length(ind)>0)
    {
      vmessage("Warning in '%s': Replaced %d negative distance(s) by _NaN(s).", _function_name(), length(ind));
      dist[ind] = _NaN;
    }
    variable SunGCDist = qualifier("SunGCDist", 8.40); % Model I in Irrgang et al., 2013, A&A, 549, A137
    variable vxs = qualifier("vxs", 11.1);  % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vys = qualifier("vys", 12.24); % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vzs = qualifier("vzs", 7.25);  % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vlsr = qualifier("vlsr", 242.0); % Model I in Irrgang et al., 2013, A&A, 549, A137
    % convert right ascension and declination to radians:
    variable ra, decl;
    (ra, decl) = RD2rad(ah, am, as, dd, dm, ds);
    % convert polar coordinates to Cartesian celestial coordinates:
    variable cartCC = Array_Type[3]; % [x, y, z]
    cartCC[0] = dist*cos(ra)*cos(decl);
    cartCC[1] = dist*sin(ra)*cos(decl);
    cartCC[2] = dist*sin(decl);
    % rotation matrix from celestial coordinates to coordinates that are aligned to Galactic enter (x-axis) and north Galactic pole (z-axis)
    % see below ("if(qualifier_exists("GC_NGP")==1)") for a derivation of the matrix
    variable rotMat = Double_Type[3,3];
    % initialize rotMat with matrix derived from coordinates for Galactic center (17 45 37.224, -28 56 10.23) and north Galactic pole (12 51 26.282, 27 07 42.01)
    % as given in Reid & Brunthaler, 2004, ApJ, 616, 872
    rotMat[0,0] = -0.05487395617553902;
    rotMat[0,1] = -0.8734371822248346;
    rotMat[0,2] = -0.48383503143198114;
    rotMat[1,0] =  0.4941107750655301;
    rotMat[1,1] = -0.44482861498054843;
    rotMat[1,2] =  0.7469819577865507;
    rotMat[2,0] = -0.8676654832928334;
    rotMat[2,1] = -0.1980782471692138;
    rotMat[2,2] =  0.45598422900423397;
    if(qualifier_exists("GC_NGP")==1) % use other coordinates for Galactic center and north Galactic pole -> re-calculate rotMat
    {
      variable a_GC, d_GC, a_NGP, d_NGP; % right ascension (a) and declination (d) of Galactic center (GC, x-axis) and north Galactic pole (NGP, z-axis)
      (a_GC, d_GC) = RD2rad( qualifier("GC_NGP")[0],  qualifier("GC_NGP")[1],  qualifier("GC_NGP")[2],
			     qualifier("GC_NGP")[3],  qualifier("GC_NGP")[4],  qualifier("GC_NGP")[5] );
      (a_NGP, d_NGP) = RD2rad( qualifier("GC_NGP")[6],  qualifier("GC_NGP")[7],  qualifier("GC_NGP")[8],
			       qualifier("GC_NGP")[9],  qualifier("GC_NGP")[10],  qualifier("GC_NGP")[11] );
      variable v_GC = [cos(a_GC)*cos(d_GC), sin(a_GC)*cos(d_GC), sin(d_GC)]; % x-axis
      variable v_NGP = [cos(a_NGP)*cos(d_NGP), sin(a_NGP)*cos(d_NGP), sin(d_NGP)]; % z-axis
      % v_GC and v_NGP are not totally orthogonal -> subtract projection of v_NGP on v_GC from v_NGP to make them orthogonal effectively changing the coordinates of
      % the north Galactic pole; however, effects ( (12 51 26.282, 27 07 42.01) -> (12 51 26.305713809790063, 27 07 41.80494829525571) ) are totally negligible to
      % uncertainty being approximately 0.1 degree (see Reid & Brunthaler, 2004, ApJ, 616, 872)
      v_NGP = v_NGP - sum(v_GC*v_NGP)*v_GC;
      v_NGP = v_NGP/sqrt(sum(v_NGP^2)); % normalization
      variable v_y = -[ v_GC[1]*v_NGP[2]-v_GC[2]*v_NGP[1], v_GC[2]*v_NGP[0]-v_GC[0]*v_NGP[2], v_GC[0]*v_NGP[1]-v_GC[1]*v_NGP[0]]; % y-axis
      rotMat[0,0] = v_GC[0];
      rotMat[0,1] = v_GC[1];
      rotMat[0,2] = v_GC[2];
      rotMat[1,0] = v_y[0];
      rotMat[1,1] = v_y[1];
      rotMat[1,2] = v_y[2];
      rotMat[2,0] = v_NGP[0];
      rotMat[2,1] = v_NGP[1];
      rotMat[2,2] = v_NGP[2];
    }
    % multiply the Celestial Cartesian data with the rotation matrix
    % to get the coordinates in Galactic system and add shift for the
    % sun-gc distance:
    variable x, y, z;
    x = rotMat[0,0]*cartCC[0] + rotMat[0,1]*cartCC[1] + rotMat[0,2]*cartCC[2] - SunGCDist;
    y = rotMat[1,0]*cartCC[0] + rotMat[1,1]*cartCC[1] + rotMat[1,2]*cartCC[2];
    z = rotMat[2,0]*cartCC[0] + rotMat[2,1]*cartCC[1] + rotMat[2,2]*cartCC[2];
    % velocity calculations:
    % convert dist from kpc to km, and SunGCDist from kpc to km:
    variable kpc_to_km = 3.08567802e+16; % conversion factor from kpc to km
    dist *= kpc_to_km;
    SunGCDist *= kpc_to_km;
    % convert pmd and pma from milliarcseconds/year to radians/s
    variable mas_yr_to_rad_s = 1.536282790e-16; % conversion factor from mas/yr to radians/s = 2.*PI/(360.*60.*60.)/1000./(365.24977652625375*24.*60.*60.);
    pmd *= mas_yr_to_rad_s;
    pma *= mas_yr_to_rad_s;
    % convert radial velocity to a Cartesian vector in Celestial coordinates
    variable radvCC = Array_Type[3]; % [vx, vy, vz]
    radvCC[0] = rv*cos(ra)*cos(decl);
    radvCC[1] = rv*sin(ra)*cos(decl);
    radvCC[2] = rv*sin(decl);
    % convert angular declination velocity to a Cartesian vector in Celestial coordinates
    variable dvCC = Array_Type[3]; % [vx, vy, vz]
    dvCC[0] = -pmd*dist*cos(ra)*sin(decl);
    dvCC[1] = -pmd*dist*sin(ra)*sin(decl);
    dvCC[2] =  pmd*dist*cos(decl);
    % convert angular right ascention velocity to a Cartesian vector in Celestial coordinates
    variable avCC = Array_Type[3]; %[vx, vy, vz]
    avCC[0] = -pma*dist*sin(ra); % factor cos(decl) already in pma
    avCC[1] =  pma*dist*cos(ra); % factor cos(decl) already in pma
    avCC[2] =  dist*0.;
    % sum the three velocity components:
    variable totalvCC = Array_Type[3]; % [vx, vy, vz]
    totalvCC[0] = radvCC[0] + dvCC[0] + avCC[0];
    totalvCC[1] = radvCC[1] + dvCC[1] + avCC[1];
    totalvCC[2] = radvCC[2] + dvCC[2] + avCC[2];
    % rotate the vector by the rotation matrix, and add the motions of the Sun within the local standard of rest (vxs, vys, vzs),
    % and the motion of the local standard of rest itself (vlsr):
    variable vx, vy, vz;
    vx = rotMat[0,0]*totalvCC[0] + rotMat[0,1]*totalvCC[1] + rotMat[0,2]*totalvCC[2] + vxs;
    vy = rotMat[1,0]*totalvCC[0] + rotMat[1,1]*totalvCC[1] + rotMat[1,2]*totalvCC[2] + vys + vlsr;
    vz = rotMat[2,0]*totalvCC[0] + rotMat[2,1]*totalvCC[1] + rotMat[2,2]*totalvCC[2] + vzs;
    if(length(x)==1) {return (x[0], y[0], z[0], vx[0], vy[0], vz[0]);}
    else return (x, y, z, vx, vy, vz);
  }
}%}}}

define gal2cel() %{{{
%!%+
%\function{gal2cel}
%\synopsis{Transform Cartesian Galactic to celestial coordinates}
%\usage{gal2cel(Double_Types x[], y[], z[], vx[], vy[], vz[]; qualifiers)}
%\description
%    Transform right-handed, Cartesian Galactic coordinates (x [kpc], y [kpc], z [kpc], vx [km/s],
%    vy [km/s], vz [km/s]) with the Galactic center at the origin, the Sun on the negative x-axis
%    and the z-axis pointing to the north Galactic pole implying clockwise Galactic rotation when
%    seen from the half space with positive z to celestial coordinates (right ascension [h, m, s],
%    declination [deg, arcmin, arcsec], distance [kpc], radial velocity [km/s], proper motion in
%    right ascension times cosine of declination [mas/yr], proper motion in declination [mas/yr]).
%\qualifiers{
%\qualifier{SunGCDist}{Sun-Galactic center distance [kpc];
%      default: 8.4 (see Model I in Irrgang et al., 2013, A&A, 549, A137)}
%\qualifier{vxs}{Sun's x-velocity component [km/s] relative to the local standard of rest;
%      default: 11.1 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vys}{Sun's y-velocity component [km/s] relative to the local standard of rest;
%      default: 12.24 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vzs}{Sun's z-velocity component [km/s] relative to the local standard of rest;
%      default: 7.25 (Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829)}
%\qualifier{vlsr}{Local standard of rest velocity [km/s];
%      default: 242 (see Model I in Irrgang et al., 2013, A&A, 549, A137)}
%\qualifier{GC_NGP}{Celestial coordinates of the Galactic center and the north Galactic pole in the
%      format [ah_GC, am_GC, as_GC, dd_GC, dm_GC, ds_GC, ah_NGP, am_NGP, as_NGP, dd_NGP, dm_NGP, ds_NGP];
%      default: [17, 45, 37.224, -28, 56, 10.23, 12, 51, 26.282, 27, 07, 42.01] (Reid & Brunthaler, 2004, ApJ, 616, 872)}
%}
%\example
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel(-8.4, 0, 0, 11.1, 254.24, 7.25);
%	% -> Sun's position and velocity
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel(0, 0, 0, 0, 0, 0);
%	% -> Galactic center's position and velocity
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel([-8.4,0], [0,0], [0,0], [11.1,0], [254.24,0], [7.25,0]);
%	% -> position and velocity of Sun and Galactic center stored in the same array
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel(-8.4, 0, 0, 11.1, 254.24, 7.25;
%	GC_NGP=[17, 45, 37.1991, -28, 56, 10.221, 12, 51, 26.2755, 27, 7, 41.704]);
%	% -> use non-standard coordinates for Galactic center and north Galactic pole
%	(ah, am, as, dd, dm, ds, dis, vrad, pma, pmd) = gal2cel( cel2gal(17, 45, 37.224, -28, 56, 10.23, 8.33, -11.1, -2.91, -5.12) );
%\seealso{cel2gal, rad2RD}
%!%-
{
  if(_NARGS!=6)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  else
  {
    variable x, y, z, vx, vy, vz;
    (x, y, z, vx, vy, vz) = ();
    x = [1.*x]; y = [1.*y]; z = [1.*z]; vx = [1.*vx]; vy = [1.*vy]; vz = [1.*vz];
    variable SunGCDist = qualifier("SunGCDist", 8.40); % see Model I in Irrgang et al., 2013, A&A, 549, A137
    variable vxs = qualifier("vxs", 11.1);  % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vys = qualifier("vys", 12.24); % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vzs = qualifier("vzs", 7.25);  % from Schoenrich, Binney & Dehnen, 2010: MNRAS 403, 1829-1833
    variable vlsr = qualifier("vlsr", 242.0);  % see Model I in Irrgang et al., 2013, A&A, 549, A137
    % shift Cartesian coordinate system from Galactic center to Sun:
    x += SunGCDist;
    vx -= vxs;
    vy -= (vys + vlsr);
    vz -= vzs;
    % rotation matrix from Galactically aligned to celestially aligned Cartesian coordinates
    % see below ("if(qualifier_exists("GC_NGP")==1)") for a derivation
    variable rotMat = Double_Type[3,3];
    % initialize rotMat with matrix derived from coordinates for Galactic center (17 45 37.224, -28 56 10.23) and north Galactic pole (12 51 26.282, 27 07 42.01)
    % as given in Reid & Brunthaler, 2004, ApJ, 616, 872
    rotMat[0,0] = -0.05487395617553902;
    rotMat[1,0] = -0.8734371822248346;
    rotMat[2,0] = -0.48383503143198114;
    rotMat[0,1] =  0.4941107750655301;
    rotMat[1,1] = -0.44482861498054843;
    rotMat[2,1] =  0.7469819577865507;
    rotMat[0,2] = -0.8676654832928334;
    rotMat[1,2] = -0.1980782471692138;
    rotMat[2,2] =  0.45598422900423397;
    if(qualifier_exists("GC_NGP")==1) % use other coordinates for Galactic center and north Galactic pole -> re-calculate rotMat
    {
      variable a_GC, d_GC, a_NGP, d_NGP; % right ascension (a) and declination (d) of Galactic center (GC, x-axis) and north Galactic pole (NGP, z-axis)
      (a_GC, d_GC) = RD2rad( qualifier("GC_NGP")[0],  qualifier("GC_NGP")[1],  qualifier("GC_NGP")[2],
			     qualifier("GC_NGP")[3],  qualifier("GC_NGP")[4],  qualifier("GC_NGP")[5] );
      (a_NGP, d_NGP) = RD2rad( qualifier("GC_NGP")[6],  qualifier("GC_NGP")[7],  qualifier("GC_NGP")[8],
			       qualifier("GC_NGP")[9],  qualifier("GC_NGP")[10],  qualifier("GC_NGP")[11] );
      variable v_GC = [cos(a_GC)*cos(d_GC), sin(a_GC)*cos(d_GC), sin(d_GC)]; % x-axis
      variable v_NGP = [cos(a_NGP)*cos(d_NGP), sin(a_NGP)*cos(d_NGP), sin(d_NGP)]; % z-axis
      % v_GC and v_NGP are not totally orthogonal -> subtract projection of v_NGP on v_GC from v_NGP to make them orthogonal effectively changing the coordinates of
      % the north Galactic pole; however, effects ( (12 51 26.282, 27 07 42.01) -> (12 51 26.305713809790063, 27 07 41.80494829525571) ) are totally negligible to
      % uncertainty being approximately 0.1 degree (see Reid & Brunthaler, 2004, ApJ, 616, 872)
      v_NGP = v_NGP - sum(v_GC*v_NGP)*v_GC;
      v_NGP = v_NGP/sqrt(sum(v_NGP^2)); % normalization
      variable v_y = -[ v_GC[1]*v_NGP[2]-v_GC[2]*v_NGP[1], v_GC[2]*v_NGP[0]-v_GC[0]*v_NGP[2], v_GC[0]*v_NGP[1]-v_GC[1]*v_NGP[0]]; % y-axis
      rotMat[0,0] = v_GC[0];
      rotMat[1,0] = v_GC[1];
      rotMat[2,0] = v_GC[2];
      rotMat[0,1] = v_y[0];
      rotMat[1,1] = v_y[1];
      rotMat[2,1] = v_y[2];
      rotMat[0,2] = v_NGP[0];
      rotMat[1,2] = v_NGP[1];
      rotMat[2,2] = v_NGP[2];
    }
    % rotate Cartesian coordiante system so that it is aligned to right ascension and declination:
    variable cartCC = Array_Type[3]; % Cartesian celestial coordinates
    cartCC[0] = rotMat[0,0]*x + rotMat[0,1]*y + rotMat[0,2]*z;
    cartCC[1] = rotMat[1,0]*x + rotMat[1,1]*y + rotMat[1,2]*z;
    cartCC[2] = rotMat[2,0]*x + rotMat[2,1]*y + rotMat[2,2]*z;
    variable cartCV = Array_Type[3]; % Cartesian celestial velocities
    cartCV[0] = rotMat[0,0]*vx + rotMat[0,1]*vy + rotMat[0,2]*vz;
    cartCV[1] = rotMat[1,0]*vx + rotMat[1,1]*vy + rotMat[1,2]*vz;
    cartCV[2] = rotMat[2,0]*vx + rotMat[2,1]*vy + rotMat[2,2]*vz;
    variable kpc_to_km = 3.08567802e+16; % conversion factor from kpc to km
    cartCC[0] *= kpc_to_km;
    cartCC[1] *= kpc_to_km;
    cartCC[2] *= kpc_to_km;
    variable dis, ra, theta, vrad, pma, pmd;
    % transform Cartesian to spherical coordinates:
    (dis, ra, theta, vrad, pma, pmd) = cart2sphere(cartCC[0], cartCC[1], cartCC[2], cartCV[0], cartCV[1], cartCV[2]);
    variable decl = -theta + PI/2.; % due to different definitions of azimuth angle theta and declination delta
    pmd *= -1.; % due to different definitions of azimuth angle theta and declination delta
    variable ah, am, as, dd, dm, ds;
    (ah, am, as, dd, dm, ds) = rad2RD(ra,decl);
    variable mas_yr_to_rad_s = 1.536282790e-16; % conversion factor from mas/yr to radians/s = 2.*PI/(360.*60.*60.)/1000./(365.24977652625375*24.*60.*60.);
    pmd /= mas_yr_to_rad_s;
    pma /= mas_yr_to_rad_s;
    pma *= cos(decl); % we use pma*cos(decl) instead of pma alone
    dis /= kpc_to_km;
    % return output:
    if(length(dis)==1) {return (ah[0], am[0], as[0], dd[0], dm[0], ds[0], dis[0], vrad[0], pma[0], pmd[0]);}
    else return (ah, am, as, dd, dm, ds, dis, vrad, pma, pmd);
  }
}%}}}
