private define Lane_Emden_eom(t,y,n) %{{{
{
  return [y[1], -(y[0])^n-2./t*y[1]]; % the Lane-Emden equation written as a system of first-order differential equations
}%}}}

define solve_Lane_Emden_equation() %{{{
%!%+
%\function{solve_Lane_Emden_equation}
%\synopsis{Solve the Lane-Emden equation numerically}
%\usage{Struct_Type s = solve_Lane_Emden_equation(Double_Type n)}
%\description
%    The Lane-Emden equation for index 0<'n'<5 is a second order
%    differential equation of the form
%
%         1   d (     dtheta)
%      ---- --- (xi^2 ------) = -theta^n
%      xi^2 dxi (        dxi)
%
%    with the boundary conditions
%
%                 dtheta
%      theta = 1, ------ = 0 at xi = 0 .
%                    dxi
%
%    The solution 'theta' satisfying these boundary conditions
%    is called Lane-Emden function of index 'n' and is returned
%    by this function together with its derivative dtheta.
%
%    The Lane-Emden functions are used to study the structures
%    of so-called polytropes, that is, gaseous spheres in hydro-
%    static equilibrium which obey a polytropic equation of state:
%      pressure = constant times density^((n+1)/n).
%    For details, see, e.g., Chapter 2.4 in the book "Principles
%    of Stellar Evolution and Nucleosynthesis" by D.D. Clayton.
%    The output structure contains also the polytropic constants
%      Rn = first zero of the Lane-Emden function of index n,
%
%                 dtheta
%      Mn = -xi^2 ------ evaluated at Rn.
%                    dxi
%
%            ( 3 dtheta)^(-1)
%      Dn = -(-- ------)      evaluated at Rn,
%            (xi    dxi)
%
%           Rn^((n-3)/n)(3Dn)^((3-n)/3n)
%      Bn = ---------------------------- .
%                (n+1)Mn^((n-1)/n)
%\notes
%    The numerical solution of the Lane-Emden equation is achieved
%    here by rewritting it as a system of first order differential
%    equations
%           d
%         --- [y_0,y_1] = [y_1, -(y_0)^n - 2/xi y_1 ]
%         dxi
%    which then allows an adaptive Runge-Kutta method of 4./5. order
%    to be applied.
%\qualifiers{
%\qualifier{tol}{[=1e-10] Absolute error control tolerance. Lower limit is 1e-15.}
%}
%\example
%    s = solve_Lane_Emden_equation(3);
%    print(s);
%\seealso{polytropic_standard_model}
%!%-
{
  if(_NARGS!=1)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable n = ();
  ifnot(0<n<5)
    throw UsageError, sprintf("Usage error in '%s': The polytropic index 'n' has to fulfill 0<n<5.", _function_name());
  variable tol = qualifier("tol", 1e-10);                                           % numerical tolerance
  variable dt = 1e-5;                                                               % initial stepsize
  variable t = 1e-5;                                                                % independent variable, initialized slightly off-center
  variable y = [1-1./6.*t^2+n/120.*t^4, -1./3.*t+n*4./120.*t^3];                    % initial condition, see Problem 2-56 in D.D. Claytons's book "Principles of Stellar Evolution and Nucleosynthesis"
  variable s = struct{ xi={0,t}, theta={1,y[0]}, dtheta={0,y[1]}, Rn, Mn, Dn, Bn }; % bookkeeping of results
  %
  variable method = qualifier("method", "RKCK");
  variable verbose = qualifier_exists("verbose");
  % ============================
  % initializing the ODE-solver:
  variable k1, k2, k3, k4, k5, k6, k7, dy_c, dy_e; % dy_c = dy continue, dy_e = dx error estimation
  variable coef = Double_Type[7,7]; % modified Butcher tableau for Runge-Kutta Cash-Karp method
  coef[0,0]= 1./5;  coef[0,1]= 1./5;
  coef[1,0]= 3./10; coef[1,1]= 3./40;       coef[1,2]= 9./40;
  coef[2,0]= 3./5;  coef[2,1]= 3./10;       coef[2,2]= -9./10;       coef[2,3]= 6./5;
  coef[3,0]= 1.;    coef[3,1]= -11./54;     coef[3,2]= 5./2;         coef[3,3]= -70./27;       coef[3,4]= 35./27;
  coef[4,0]= 7./8;  coef[4,1]= 1631./55296; coef[4,2]= 175./512;     coef[4,3]= 575./13824;    coef[4,4]= 44275./110592; coef[4,5]= 253./4096;
  coef[5,0]= 0;     coef[5,1]= 2825./27648; coef[5,2]= 18575./48384; coef[5,3]= 13525./55296;  coef[5,4]= 277./14336;    coef[5,5]= 1./4; % fourth order solution
  coef[6,0]= 0;     coef[6,1]= 37./378;     coef[6,2]= 250./621;     coef[6,3]= 125./594;      coef[6,4]= 0;             coef[6,5]= 512./1771; % fitfh order solution
  % ============================
  % =================
  % solving the ODEs:
  while(y[0]>tol)
  {
    k1 = dt * Lane_Emden_eom(t, y, n);
    k2 = dt * Lane_Emden_eom(t + coef[0,0]*dt, y + coef[0,1]*k1, n);
    k3 = dt * Lane_Emden_eom(t + coef[1,0]*dt, y + coef[1,1]*k1 + coef[1,2]*k2, n);
    k4 = dt * Lane_Emden_eom(t + coef[2,0]*dt, y + coef[2,1]*k1 + coef[2,2]*k2 + coef[2,3]*k3, n);
    k5 = dt * Lane_Emden_eom(t + coef[3,0]*dt, y + coef[3,1]*k1 + coef[3,2]*k2 + coef[3,3]*k3 + coef[3,4]*k4, n);
    k6 = dt * Lane_Emden_eom(t + coef[4,0]*dt, y + coef[4,1]*k1 + coef[4,2]*k2 + coef[4,3]*k3 + coef[4,4]*k4 + coef[4,5]*k5, n);
    dy_c = coef[5,1]*k1 + coef[5,2]*k3 + coef[5,3]*k4 + coef[5,4]*k5 + coef[5,5]*k6; % fourth order solution
    dy_e = coef[6,1]*k1 + coef[6,2]*k3 + coef[6,3]*k4 + coef[6,4]*k5 + coef[6,5]*k6; % fifth order solution
    variable scale = 0.9*(tol/max(abs(dy_c-dy_e)))^(0.2); % 0.9 is safety factor; derivation see Numerical Recipes, Section "Adaptive Stepsize Control for Runge-Kutta"
    % variable scale = ( 0.5*tol*dt/max(abs(dy_c-dy_e)) )^0.25;  % "derivation of this formula can be found in advanced books on numerical analysis", statement in Numerical Methods Using Matlab, 4th Edition, 2004
    if(scale>1 or dt<=1e-15*t) % scale>1: accept stepsize only if correction factor is larger than 1
    {                          % dt<=abs(1e-15*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
      y += dy_c;
      t += dt;
      list_append(s.xi, t);
      list_append(s.theta, y[0]);
      list_append(s.dtheta, y[1]);
    }
    dt = _min(_max(dt*scale, t*1e-15), abs(y[0]/y[1])); % abs(1e-15*t) in order to avoid infinite loops due to Double_Type precision limit at dt=1e-16*t
  }                                                     % abs(y[0]/y[1]) is based on Newton's method to find the root of a function and is thus important when theta approaches zero
  s.xi = list_to_array(s.xi, Double_Type);
  s.theta = list_to_array(s.theta, Double_Type);
  s.dtheta = list_to_array(s.dtheta, Double_Type);
  s.Rn = s.xi[-1];
  s.Mn = -s.Rn^2*s.dtheta[-1];
  s.Dn = 1./(-3/s.Rn*s.dtheta[-1]);
  s.Bn = s.Rn^((n-3.)/n)*(3.*s.Dn)^((3.-n)/3./n)/(n+1.)/s.Mn^((n-1.)/n);
  return s;
}%}}}
