define DateOfJD() {
%!%+
%\function{DateOfJD}
%\synopsis{Calculates the date corresponding to a JD value}
%\usage{Struct_Type date = DateOfJD(JD);}
%\description
%    The return value is a structure of the following form:\n
%       \code{date = struct { year, month, day, hour, minute, second };}
%    The algorithm used is that by Meeus, Astronomical Formulae for
%    Calculators. This routine is array safe.
%\qualifiers{
%\qualifier{julianswitch}{For JD smaller than this date, return the date
%        in the Julian calendar rather than in the Gregorian
%        calendar. The default is JD2299161 (15 October 1582),
%        which is appropriate for most of catholic Europe.
%        Note, however, that there are countries that switched
%        only much later (see https://en.wikipedia.org/wiki/Julian_calendar
%        and https://de.wikipedia.org/wiki/Gregorianischer_Kalender).    
%        For example, Russia only switched on 1 February 1918 (Gregorian).}
%\qualifier{julian_calendar}{force returning the date in the Julian Calendar
%        (ignore julianswitch)}
%\qualifier{gregorian_calendar}{force returning the date in the Gregorian Calendar
%        (ignore julianswitch)}
%}
%\seealso{JDofDate, MJDofDate}
%!%-
    variable JD;
    switch(_NARGS)
    { case 1: JD=(); }
    { help(_function_name()); return;}

    variable ndx=where(JD<0.);
    if (length(ndx)>0) {
	throw UsageError,sprintf("%s: only positive JD are allowed\n",_function_name());
    }

    variable retval=struct{year,month,day,hour,minute,second};

    if (qualifier_exists("julian_calendar") or qualifier_exists("gregorian_calendar")) {
	if (qualifier_exists("julianswitch")) {
	    throw UsageError,sprintf("%s: julianswitch and julian_calendar or gregorian_calendar qualifiers are not allowed at the same time\n",_function_name());
	}
    }

    if (qualifier_exists("julian_calendar") and qualifier_exists("gregorian_calendar")) {
	    throw UsageError,sprintf("%s: julian_calendar or gregorian_calendar qualifiers are mutually exclusive\n",_function_name());
    }

    variable jswitch=qualifier("julianswitch",2299161);
    if (qualifier_exists("gregorian_calendar")) {
	jswitch=-1.; % force all dates to gregorian (since JD<0 is not allowed)
    }
    if (qualifier_exists("julian_calendar")) {
	jswitch=1.+max(JD); % force all dates to julian
    }
    
    
    %
    % implement the Algorithm of Meeus
    % Note that this algorithm depends crucially on the peculiarities
    % of floating point arithmetics vs. integer arithmetics, so please
    % change the following lines only if you REALLY know what you are
    % doing and have a copy of Meeus next to you.
    %
    variable t=JD+0.5;
    variable z=int(t);
    variable f=t-z;
    variable a;
    variable x;

    if (length(JD)>1) {
	variable ndx1,ndx2;
	ndx1=where(JD<jswitch,&ndx2);
	a=Double_Type[length(JD)];
	if (length(ndx1)>0) {
	    a[ndx1]=z[ndx1];
	}
	if (length(ndx2)>0) {
	    x=int((z[ndx2]-1867216.25)/36524.25);
	    a[ndx2]=z[ndx2]+1.+x-int(x/4.);
	}
    } else {
	if (JD<jswitch) {
	    a=z;
	} else {
	    x=int((z-1867216.25)/36524.25);
	    a=z+1.+x-int(x/4.);
	}
    }
    variable b=a+1524.;
    variable c=int((b-122.1)/365.25);
    variable d=int(365.25*c);
    variable e=int((b-d)/30.6001);
    retval.day=int(b-d-int(30.6001*e));
    if (length(JD)>1) {
	retval.month=Integer_Type[length(JD)];
	retval.year=Integer_Type[length(JD)];
	ndx1=where(e<13.5,&ndx2);
	if (length(ndx1)>0) {
	    retval.month[ndx1]=int(e[ndx1]-1.);
	}
	if (length(ndx2)>0) {
	    retval.month[ndx2]=int(e[ndx2]-13.);
	}
	retval.year=int(c-4715);
	ndx1=where(retval.month>=3);
	if (length(ndx1)>0) {
	    retval.year[ndx1]=retval.year[ndx1]-1;
	}
    } else {
	retval.month=int ( (e<13.5) ? e-1. : e-13. );
	retval.year=int ( (retval.month<3) ? c-4715. : c-4716.);
    }
    f=f*24.;
    retval.hour=int(f);
    f=(f-retval.hour)*60.;
    retval.minute=int(f);
    retval.second=(f-retval.minute)*60.;
    return retval;
}
