define GMST()
%!%+
%\function{GMST}
%\synopsis{Calculates the Greenwich sidereal time and the local siderial time}
%\usage{Double_Type GST = GMST(Double_Type JD);}
%
%\qualifiers{
%\qualifier{hour}{Return the GMST in hours, not in seconds.}
%\qualifier{dut1}{The difference UT1-UTC in seconds (see IERS Bull. A).}
%\qualifier{mjd}{Argument is in MJD, not in JD.}
%\qualifier{gast}{Return the Greenwhich Apparent Siderial Time, GAST.}
%\qualifier{era}{Return the Earth rotation angle (in rotations).}
%\qualifier{lst}{Return the local mean siderial time LST.}
%\qualifier{lon}{longitude for the calculation of the LST (east is positive;
%               in geodetic [ITRS] coordinates, for
%               practical purposes these are identical to the
%               WGS-84 coordinates given by GPS).}
%\qualifier{lat}{latitude for the calculation of the LST (north is positive,
%                  only needed if xp and yp are given)}
%\qualifier{deg}{lon and lat are in degrees, not in radian.}
%\qualifier{xp}{x-coordinate of the Celestial Intermediate Pole (CIP) 
%               ALWAYS in arcseconds, as given by the IERS bulletin.}
%\qualifier{yp}{y-coordinate of the CIP, ALWAYS in arcseconds.}
%}
%\description
%    This code calculates the Greenwich Mean Siderial Time and the
%    Greenwhich Apparent Siderial Time in seconds.
%
%    The argument \code{JD} is the Julian Date in the UT1 time system.
%    For most applications, one can take UT1=UTC. For the most precise
%    applications, a correction term  DUT1=UT1-UTC can be looked up in
%    IERS Bulletin A
%    (see https://www.iers.org/IERS/EN/Publications/Bulletins/bulletins.html).
%
%    The algorithms used are described by Kaplan (2005, USNO Circular
%    179, Sect. 2.6.2).
%
%    This routine is array safe.
%
%    Note that contrary to the coordinate transform routines the
%    deg qualifier only affects the interpretation of the lon, lat
%    qualifiers and does NOT affect the other arguments or return
%    arguments!
%\seealso{equation_equinoxes}
%!%-
{
  variable JD;
  switch(_NARGS)
  { case 1: JD = (); }
  { help(_function_name()); return; }

  if (qualifier_exists("mjd")) {
      JD=mjd2jd(JD);
  }
  JD+=qualifier("dut1",0.)/86400.;

  variable fra=JD mod 1.;
  variable du=JD-2451545.0;
  
  % Kaplan, eq. 2.11.
  % Do NOT change the parentheses!
  variable era=(0.7790572732640+fra)+0.00273781191135448*du;

  if (qualifier_exists("era")) {
      return era;
  }


  variable T;
  if (typeof(JD)==Array_Type) {
      T=Double_Type[length(JD)];
      variable ndx1;
      variable ndx=where(JD>=2443144.5,&ndx1);
      if (length(ndx)>0) {
	  T[ndx]=(tai2tt(utc2tai(JD[ndx]))-2451545.0)/36525.;
      }
      %FIXME: we're assuming that ut=tt before 1977 Jan 1.
      %this is not entirely correct, check Meeus here!
      if (length(ndx1)>0) {
	  T[ndx1]=(tai2tt(JD[ndx1])-2451545.0)/36525.;
      }
  } else {
      if (JD>=2443144.5) {
	  T=(tai2tt(utc2tai(JD))-2451545.0)/36525.;
      } else {
	  T=(tai2tt(JD)-2451545.0)/36525.;
      }
  }
  
  % Kaplan, eq. 2.12
  variable gmst=86400.0*era+
                (((((-0.0000000368*T-0.000029956)*T
                -0.00000044)*T+1.3915817)*T+4612.156534)*T+0.014506)/15.;

  if (qualifier_exists("lst")) {
      variable lon=qualifier("lon",0.);
      variable lat=qualifier("lat",0.);
      if (qualifier_exists("deg")) {
	  lon*=PI/180.;
	  lat*=PI/180.;
      }

      variable xp=qualifier("xp",0.);
      variable yp=qualifier("yp",0.);
      if (xp!=0. or yp!=0.) {
	  variable sl,cl;
	  (sl,cl)=sincos(lon);
	  lon+=(xp*sl+yp*cl)*tan(lat)/3600.;
      }
      % ...equation for longitude in rad
      % gmst=gmst+(3600./15.)*lon*180./PI;
      gmst=gmst+43200*lon/PI;
  }

  if (qualifier_exists("gast")) {
      variable equ=equation_equinoxes(JD;seconds);
      gmst+=equ;
  }

  
  gmst=pos_modulo(gmst,86400.);

  if(qualifier_exists("hour")) {
      return gmst/3600.;
  }
  return gmst;
}
