define JDofEpoch()
%!%+
%\function{JDofEpoch}
%\synopsis{Returns the (M)JD of a Julian or Besselian epoch such as J2000.0}
%\usage{jd=JDofEpoch(epochstring;qualifiers)}
%\qualifiers{
%\qualifier{mjd}{Return the modified julian date (default: JD)}
%}
%\description
%   This function returns the julian date corresponding to a
%   Julian or Besselian epoch (e.g., J2000, B1950.0 etc.). For highest
%   precision, the most commonly used values are hardcoded, other values
%   are calculated using the formulae given by Lieske (1979, A&A 73, 282)
%   for Besselian dates and the standard definition of the Julian epoch
%   (based on Julian years with 365.25 days).
%
%   The epoch string must start with a "B" for a Besselian epoch and
%   "J" for a Julian epoch, followed by a number that designates the year
%   fractional values such as 2017.25 are possible.
%
%   Note that the time system in which the dates are defined is TT, for most
%   applications that do not require microsecond accuracy for the conversion
%   of positions the precise time system won't matter.
%
%   If the routine is called with a floating point number, it is assumed
%   that this number is the epoch and this number is returned. This greatly
%   simplifies functions in which the epoch is either a string or
%   a number.
%
%\seealso{EpochofJD,tt2tai,tt2tdb}
%!%-
{
    variable ep;

    switch(_NARGS)
    { case 1: ep=(); }
    { help(_function_name());return;}

    % if called with a number: do nothing
    if (__is_numeric(ep)) {
	return typecast(ep,Double_Type);
    }

    % deal with the most common cases first
    if (qualifier_exists("mjd")) {
	switch(ep)
	{ case "J2000.0": return 51544.5;}
	{ case "B1950.0": return 33281.9235;}
	{ case "B1900.0": return 15019.81352;}
    } else {
	switch(ep)
	{ case "J2000.0": return 2451545.0;}
	{ case "B1950.0": return 2433282.4235;}
	{ case "B1900.0": return 2415020.31352;}
    }
    
    %
    % general epoch string -> decode the string
    variable type=substr(ep,1,1);
    variable year=atof(substr(ep,2,strlen(ep)));

    if(type!="J" && type !="B") {
	throw UsageError,sprintf("%s: epoch must start with J or B\n",_function_name());
    }
    
    if(type=="J") {
	% Julian epoch
	if (qualifier_exists("mjd")) {
	    return (year-2000.0)*365.25+51544.5;
	} else {
	    return (year-2000.0)*365.25+2451545.0;
	}
    } else {
	if (qualifier_exists("mjd")) {
	    return (year-1900.) * 365.242198781 + 15019.81352;
	} else {
	    return (year-1900.) * 365.242198781 + 2415020.31352;
	}
    }
}
